/*
*  Name:      exceptions.cc
*  Author:    Rafael Jesus Alcantara Perez
*  Summary:   UESQLC exceptions
*  Date:      $Date: 1999/06/26 00:26:59 $
*  Revision:  $Revision: 1.6 $
*
*  Copyright (C) 1996-1999  Rafael Jesus Alcantara Perez <ralcan@vnet.es>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
*  MA 02111-1307, USA.
*/

#include "exceptions.hh"


//
//  L O C A L   V A R I A B L E S
//

///  Warning message.
static const char*   _pkcWarningDescription = "";

///  Error message.
static const char*   _pkcErrorDescription   = "";

///  Fatal error message.
static const char*   _pkcFatalDescription   = "uesqlc: ";


//
//  C O N S T R U C T O R S
//

TExceptionUESQLC_Warning::TExceptionUESQLC_Warning ( const char* pkcSPECIFIC_DESCRIPTION ,
                                                     const char* pkcFILE_NAME            ,
                                                     const int   kiLINE_NUMBER           ) :
  TBaseException    ( _pkcWarningDescription  ,
                      pkcSPECIFIC_DESCRIPTION ,
                      pkcFILE_NAME            ,
                      kiLINE_NUMBER           )                                            ,
  TExceptionUESQLC  ()                                                                     ,
  TWarningException ()                                                                     {}


TExceptionUESQLC_Error::TExceptionUESQLC_Error ( const char* pkcSPECIFIC_DESCRIPTION ,
                                                 const char* pkcFILE_NAME            ,
                                                 const int   kiLINE_NUMBER           ) :
  TBaseException   ( _pkcErrorDescription    ,
                     pkcSPECIFIC_DESCRIPTION ,
                     pkcFILE_NAME            ,
                     kiLINE_NUMBER           )                                         ,
  TExceptionUESQLC ()                                                                  ,
  TErrorException  ()                                                                  {}


TExceptionUESQLC_Fatal::TExceptionUESQLC_Fatal ( const char* pkcSPECIFIC_DESCRIPTION ,
                                                 const char* pkcFILE_NAME            ,
                                                 const int   kiLINE_NUMBER           ) :
  TBaseException   ( _pkcFatalDescription    ,
                     pkcSPECIFIC_DESCRIPTION ,
                     pkcFILE_NAME            ,
                     kiLINE_NUMBER           )                                         ,
  TExceptionUESQLC ()                                                                  ,
  TFatalException  ()                                                                  {}


//
//  S E L E C T O R S
//

void TExceptionUESQLC::putOnStream (ostream& rtTARGET_OSTREAM) const
{

  if ( fileName() )
  {
    TBaseException::putOnStream (rtTARGET_OSTREAM);
  }
  else
  {
    rtTARGET_OSTREAM << generalDescription();
    if ( specificDescription() )
    {
      rtTARGET_OSTREAM << specificDescription();
    }
    rtTARGET_OSTREAM << endl;
  }
 
}  // putOnStream()


void TExceptionUESQLC_Fatal::putOnStream (ostream& rtTARGET_OSTREAM) const
{

  //
  //  Fatal errors doesn't depend on files.
  //
  rtTARGET_OSTREAM << generalDescription();
  if ( specificDescription() )
  {
    rtTARGET_OSTREAM << specificDescription();
  }
  rtTARGET_OSTREAM << endl;
 
}  // putOnStream()
