#ifndef _POSTGRESQL_LIBPQ_6_2_X__
#define _POSTGRESQL_LIBPQ_6_2_X__
#include <cstdlib>
#include <libpq-fe.h>
#ifdef _UESQL_DEBUG__
# include <mpcl/test.h>
#endif  // not _UESQL_DEBUG
#include <mpcl/string.hh>
#include <mpcl/extended_map.hh>
#include <uesqlc/defs-uesqlc.hh>
#include "uesql_base_manager.hh"


//
//  UESQLC::postgresql-libpq-6.2.x internal definitions.
//

#define PG62X_INSTANTIATE(pkcFORMAT, zINDEX, ppkcVARIABLES)  \
  Format (pkcFORMAT, ppkcVARIABLES [zINDEX + 1])

#define PG62X_NOT_FOUND_CONDITION ( ( PQresultStatus (htPostgresql62xManager._ptPgresult) == PGRES_EMPTY_QUERY ) || \
                                    ( !PQntuples (htPostgresql62xManager._ptPgresult) )                           )


//
//  Execute command but don't display it.
//
#define PGIV_EXEC(pkcSOURCE_INSTRUCTION, ptSOURCE_PGCONN)                                   \
{                                                                                           \
  htPostgresql62xManager._ptAuxPgresult = PQexec (ptSOURCE_PGCONN, pkcSOURCE_INSTRUCTION);  \
  if ( htPostgresql62xManager._ptAuxPgresult )                                              \
  {                                                                                         \
    PQclear (htPostgresql62xManager._ptAuxPgresult);                                        \
  }                                                                                         \
}


//
//  Execute command and display it.
//
#define PGV_EXEC(pkcSOURCE_INSTRUCTION, ptSOURCE_PGCONN, ERROR_ACTION)                         \
{                                                                                              \
  PG_EXEC (pkcSOURCE_INSTRUCTION, ptSOURCE_PGCONN, ERROR_ACTION);                              \
  if ( htPostgresql62xManager._ptPgresult ) { PQclear (htPostgresql62xManager._ptPgresult); }  \
}


//
//  If _UESQL_DEBUG__ has been defined, then SQL statements will be
//  printed using 'test.h' header from MPCL.
//
#ifdef _UESQL_DEBUG__
# define PG_EXEC(pkcSOURCE_INSTRUCTION, ptSOURCE_PGCONN, ERROR_ACTION)  \
{  \
  if ( !htPostgresql62xManager.initialize() ) \
  { \
    ERROR_ACTION \
  } \
  TEST_WRITE_INFO ("%s", pkcSOURCE_INSTRUCTION); \
  htPostgresql62xManager._ptPgresult = PQexec (ptSOURCE_PGCONN, pkcSOURCE_INSTRUCTION); \
  if ( !htPostgresql62xManager._ptPgresult ) \
  { \
    /* \
    *   Condition:           transaction rollback \
    *   Class subcondition:  statement completion unknown \
    */ \
    uesql::SQLCODE = -1; uesql::SQLSTATE = "40003"; \
    cerr << PQerrorMessage (htPostgresql62xManager._ptPgconn); \
    ERROR_ACTION \
  } \
  else \
  { \
    if ( ( PQresultStatus (htPostgresql62xManager._ptPgresult) == PGRES_COMMAND_OK ) || ( PQresultStatus (htPostgresql62xManager._ptPgresult) == PGRES_TUPLES_OK ) ) \
    { \
      uesql::SQLCODE = 0; uesql::SQLSTATE = "00000"; \
    } \
    else \
    { \
      /* \
      *  Condition:           transaction rollback \
      *  Class subcondition:  statement completion unknown \
      */ \
      uesql::SQLCODE = -1; uesql::SQLSTATE = "40003"; \
      cerr << PQerrorMessage (htPostgresql62xManager._ptPgconn); \
      ERROR_ACTION \
    } \
  } \
}
#else
# define PG_EXEC(pkcSOURCE_INSTRUCTION, ptSOURCE_PGCONN, ERROR_ACTION)  \
{  \
  if ( !htPostgresql62xManager.initialize() ) \
  { \
    ERROR_ACTION \
  } \
  htPostgresql62xManager._ptPgresult = PQexec (ptSOURCE_PGCONN, pkcSOURCE_INSTRUCTION); \
  if ( !htPostgresql62xManager._ptPgresult ) \
  { \
    /* \
    *   Condition:           transaction rollback \
    *   Class subcondition:  statement completion unknown \
    */ \
    uesql::SQLCODE = -1; uesql::SQLSTATE = "40003"; \
    cerr << PQerrorMessage (htPostgresql62xManager._ptPgconn); \
    ERROR_ACTION \
  } \
  else \
  { \
    if ( ( PQresultStatus (htPostgresql62xManager._ptPgresult) == PGRES_COMMAND_OK ) || ( PQresultStatus (htPostgresql62xManager._ptPgresult) == PGRES_TUPLES_OK ) ) \
    { \
      uesql::SQLCODE = 0; uesql::SQLSTATE = "00000"; \
    } \
    else \
    { \
      /* \
      *  Condition:           transaction rollback \
      *  Class subcondition:  statement completion unknown \
      */ \
      uesql::SQLCODE = -1; uesql::SQLSTATE = "40003"; \
      cerr << PQerrorMessage (htPostgresql62xManager._ptPgconn); \
      ERROR_ACTION \
    } \
  } \
}
#endif  // not _UESQL_DEBUG__


//
//  UESQLC::postgresql-libpq-6.2.x internal classes.
//

class TPostgresql62xManager : public TUesqlBaseManager
{

  public:

    static PGconn*     _ptPgconn;
    static PGresult*   _ptPgresult;
    static PGresult*   _ptAuxPgresult;
    static string      _ySqlStatement;
    static string      _yHost;
    static string      _yDatabaseName;

  
  private:

    mpcl::xextended_map<string, PGconn*>   xConnectionMap;
    bool                                      gIsThereConnection;

  
  public:

    //
    //  C O N S T R U C T O R S
    //

    TPostgresql62xManager (void) :
      TUesqlBaseManager  ()      ,
      xConnectionMap     ()      ,
      gIsThereConnection (false) {}

    int initialize (void);

    void connect (const string& rkyCONNECTION_NAME, const PGconn* pktCONNECTION);

    void disconnect (const string& rkyCONNECTION_NAME);

    void disconnectCurrent (void);

    PGconn* setCurrentConnection (const string& rkyCONNECTION_NAME);


  public:

    //
    //  S E L E C T O R S
    //

    bool existsDefaultConnection (void) const;

    bool isConnected (const string& rkyCONNECTION_NAME) const;

    PGconn* currentConnection (void) const;

    PGconn* connection (const string& rkyCONNECTION_NAME) const;

    void finish (void);

};  // class TPostgresql62xManager


//
//  E X P O R T E D   V A R I A B L E S
//
//    UESQLC::postgresql-libpq-6.2.x.
//

extern TPostgresql62xManager   htPostgresql62xManager;


#endif  // not _POSTGRESQL_LIBPQ_6_2_X__
